'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

exports.default = function (babel) {
  var t = babel.types,
      version = babel.version;

  var babel6 = _semver2.default.satisfies(version, '^6.0.0');

  return {
    name: 'babel-blade', // not required
    inherits: babel6 ? _babelPluginSyntaxJsx2.default : _pluginSyntaxJsx2.default, // for jsx/babel6-7 interop
    visitor: {
      // eslint-disable-next-line complexity
      Identifier(path) {
        handleCreateRazor(path, t);
      }
    }
  };
};

exports.handleCreateRazor = handleCreateRazor;

var _semver = require('semver');

var _semver2 = _interopRequireDefault(_semver);

var _pluginSyntaxJsx = require('@babel/plugin-syntax-jsx');

var _pluginSyntaxJsx2 = _interopRequireDefault(_pluginSyntaxJsx);

var _babelPluginSyntaxJsx = require('babel-plugin-syntax-jsx');

var _babelPluginSyntaxJsx2 = _interopRequireDefault(_babelPluginSyntaxJsx);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

// data structure
var _require = require('./dataStructures'),
    RazorData = _require.RazorData;

/****
 *
 * Discussion of babel strategy
 *
 * at the top level we give users createQuery and createFragment methods.
 * when these are called and assigned, whatever identifier they get assigned to becomes a "razor".
 *
 * 1. for each razor
 * 	for each call of the razor that is assigned, that is a blade
 * 		we call parseBlade on the blade
 *
 * 2. parseBlade
 *  only assignment targets are blades
 *    process each reference (variable declarator and standalone) at once
 *
 * 2.1 process reference
 *    lhs = process LHS
 *    rhs process RHS
 *    merge(lhs, rhs)
 *    parseBlade on all assign targets (if any)
 *
 * 2.2 process LHS
 *    get LHS (either id or objectpattern's key)
 *
 * 2.3 process RHS
 *    get RHS list
 *
 * 2.4 merge()
 *    added properties can come from either lhs or rhs. rhs taking precedence
 *      rhs.foreach - call add, return child, call add again
 *      lhs.foreach - ???
 *    if rhs and lhs has no kids, just an assign, just tag the blade and early exit
 *    else, sequence is:
 *      .add from rhs, including args
 *      at the end of rhs, if there is an alias, assign the alias too
 *      if lhs has kids, keep .adding, including alias
 *
 * 3.	once all blades are parsed, compose and insert the graphql query where the razor is referenced
 *
 * LHS cases to handle:
 * - const var1 = DATA
 * - const { var2 } = DATA
 * - const { var3, var4 } = DATA
 * - const { var5: var6  } = DATA
 * - const { var7: { var8: var9} } = DATA
 *
 * interaction cases to handle:
 * - const var1 = DATA.v2
 * - const { var2 } = DATA.v2
 * - const { var3, var4 } = DATA.v2
 * - const { var5: var6  } = DATA.v2
 * - const { var7: { var8: var9} } = DATA.v2
 *
 * RHS cases
 * - const var1 = DATA.v3
 * - const var2 = DATA.v3.var3
 * - const var4 = DATA.v3.var5
 * - const var6 = DATA.v3({ foo: 1, bar: 2})
 * - const var7 = DATA.v3({ foo: 3, bar: 4}).var8
 * - DATA.v3.var8 // no assignment!
 *
 **/

// for jsx/babel6-7 interop


function handleCreateRazor(path, t) {
  if (isCreateQuery(path) || isCreateFragment(path)) {
    // get the identifier and available args
    var identifier = getAssignTarget(path);
    var queryArgs = void 0;
    if (isCallee(path)) {
      queryArgs = getCalleeArgs(path);
    }
    // clear the reference
    path.findParent(function (ppath) {
      return ppath.isVariableDeclaration();
    }).remove();
    // traverse scope for identifier references
    var refs = path.scope.bindings[identifier].referencePaths;
    if (refs.length > 0) {
      var razorID = null;
      var fragmentType = path.parent.arguments.length ? path.parent.arguments[0].value : null;
      var queryType = isCreateFragment(path) ? 'fragment' : 'query';
      var razorData = new RazorData({
        type: queryType,
        name: identifier,
        fragmentType,
        args: queryArgs
      });
      refs.forEach(function (razor) {
        // go through all razors
        if (isCallee(razor)) {
          // we have been activated! time to make a blade!
          razorID = getAssignTarget(razor);
          // clear the reference
          if (razor.container.arguments[0]) razor.parentPath.replaceWith(razor.container.arguments[0]);else razor.parentPath.remove();
          parseBlade(razor, razorID, razorData);
        }
      });

      // insert query
      refs.forEach(function (razor) {
        if (!isObject(razor)) {
          var _razorData$print = razorData.print(),
              stringAccumulator = _razorData$print.stringAccumulator,
              litAccumulator = _razorData$print.litAccumulator;

          razor.replaceWith(t.templateLiteral(stringAccumulator.map(function (str) {
            return t.templateElement({ raw: str });
          }), litAccumulator.map(function (lit) {
            return lit || t.nullLiteral();
          })));
        }
      });
    }
  }
}

function parseBlade(path, id, razorData) {
  var slice = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : 0;

  var refs = path.scope.bindings[id] && path.scope.bindings[id].referencePaths;
  // console.log('parseblade', { refs, id, razorData, path });
  if (slice) refs = refs.slice(slice);
  if (refs && refs.length > 0) {
    // there has been an assignment and it has been used
    refs.forEach(function (blade) {
      processReference(blade, razorData);
      // call parseblade on all LHS
    });
  } else {
    // there has been no assignment or it has not been used
    var bladeID = getObjectPropertyName(path);
    if (bladeID) {
      var propID = getObjectPropertyName(path);
      var child = razorData.add({ name: propID });
      parseBlade(path.parentPath, propID, child);
    }
  }
}

function processReference(blade, razorData) {
  var LHS = void 0,
      RHS = void 0;

  /***** PROCESS LHS *****/
  // for now LHS processing goes a maximum of one level.
  // will implement multilevel later

  // find declarator parent context
  var ctx = blade.findParent(function (ppath) {
    return ppath.isVariableDeclarator() && ppath.scope.uid === blade.scope.uid;
  });
  if (ctx) {
    // then there is assignment
    // if it is a top level assign, its a plain alias
    if (ctx.node.id.name) {
      LHS = ctx.node.id.name;
    } else {
      LHS = processLHS(ctx);
    }
  }

  /***** PROCESS RHS *****/

  // doing this just to read chained objects (eg foo.bar.baz)
  // from left to right instead of
  // babel AST which makes us read from right to left
  RHS = [];
  var RHSVisitor = {
    MemberExpression(childpath) {
      var aliasPath = void 0,
          args = void 0;
      if (isCallee(childpath)) {
        args = getCalleeArgs(childpath); // fancy!
        aliasPath = childpath;
      }
      RHS.push({ name: childpath.node.property.name, args, aliasPath });
      /*
        if (isCallee(childpath)) {
          args = getCalleeArgs(childpath) // fancy!
          childpath.parentPath.replaceWith(childpath) // drop the args
        } else {
        	RHS.push({name: childpath.node.property.name, args})
          	if (args) args = null // clear it if used
        }
        */
    }
  };
  if (ctx) {
    ctx.traverse(RHSVisitor);
  } else {
    // even if there is no assignment, still need to traverse the kids
    blade.findParent(function (ppath) {
      return !ppath.isMemberExpression();
    }).traverse(RHSVisitor);
  }
  RHS = RHS.reverse(); // annoying bc of how the AST works

  // console.log({ ctx, blade, LHS, RHS });

  /***** MERGE LHS AND RHS *****/

  // MERGE RHS FIRST: rhs.foreach - call add, return child, call add again
  var currentData = razorData;
  RHS.forEach(function (_ref) {
    var args = _ref.args,
        name = _ref.name,
        aliasPath = _ref.aliasPath;

    // console.log({name, args, currentData})
    currentData = currentData.add({ name, args: args }); // add fragments, directives later
    if (currentData._alias && aliasPath) aliasPath.parentPath.replaceWith(aliasPath);
    if (aliasPath) aliasPath.node.property.name = currentData._alias;
  });

  // MERGE LHS second: lhs.foreach
  if (typeof LHS === 'string') {
    // simple assignment
    parseBlade(blade.parentPath, LHS, currentData);
  } else if (Array.isArray(LHS)) {
    // there has been a destructuring!
    var LHSpointer = currentData;
    LHS.forEach(function (_ref2) {
      var name = _ref2.name,
          alias = _ref2.alias;

      LHSpointer = currentData.add({ name });
      //const temp = blade.scope.bindings[alias]
      //console.log(alias, blade, temp)
      parseBlade(blade.parentPath, alias, LHSpointer, 1);
    });
  }

  /***** DONE. below are helper functions *****/

  // helper functions for processReference

  function processLHS(ctx) {
    // if it is an object destructure, push and recurse
    var lhs = [];
    var props = ctx.node.id.properties;
    if (props) {
      props.forEach(function (prop) {
        var k = prop.key.name;
        var maybeV = prop.value.properties;
        if (maybeV) {
          // multilayers, recurse
          throw new Error('multilayer destructure not implemented yet');
        } else {
          // lhs and rhs, no multilayer
          lhs.push({
            name: k,
            alias: prop.value.name
          });
        }
      });
      return lhs; // normal terminate
    }
    // shouldnt get here
    console.log('please examine this', ctx);
  }
}

/****
 *
 * Simple readable utils for navigating the path,
 * pure functions w no significant logic
 *
 */

function getCalleeArgs(calleePath) {
  var arg = calleePath.container.arguments;
  return arg; // {first: arg[0], second: arg[1]}

  /* old attempt with direct values
  return calleePath.container.arguments
    .map(
    arg => {
      console.log({arg})
        // it is an object
      if (arg.properties) {
        const temp = {}
        arg.properties.map(
          prop => temp[prop.key.name] = prop.value.value
        )
        return temp
      } else
        return arg.value
    }
  )
  */
}

function getAssignTarget(path) {
  return path.parentPath.container.id ? path.parentPath.container.id.name : undefined;
}

function getAssignTarget2(path) {
  return path.container.id ? path.container.id.name : undefined;
}

function getObjectPropertyName(path) {
  return path.container.property ? path.container.property.name : undefined;
}

// function getFragmentName(path) {
//   if (
//     path.parentPath.isAssignmentExpression() &&
//     path.parent.left.type === 'MemberExpression' &&
//     path.parent.left.property.name === 'fragment'
//   ) {
//     const name = path.parent.left.object.name
//     return name[0].toLowerCase() + name.slice(1) + 'Fragment'
//   }
//   return null
// }

function isObject(path) {
  return looksLike(path, { key: 'object' });
}

function isCallee(path) {
  var parent = path.parentPath;
  return parent.isCallExpression() && path.node === parent.node.callee;
}

function isCreateQuery(path) {
  return looksLike(path, { node: { name: 'createQuery' } });
}
function isCreateFragment(path) {
  return looksLike(path, { node: { name: 'createFragment' } });
}

function looksLike(a, b) {
  return a && b && Object.keys(b).every(function (bKey) {
    var bVal = b[bKey];
    var aVal = a[bKey];
    if (typeof bVal === 'function') {
      return bVal(aVal);
    }
    return isPrimitive(bVal) ? bVal === aVal : looksLike(aVal, bVal);
  });
}

function isPrimitive(val) {
  // eslint-disable-next-line
  return val == null || /^[sbn]/.test(typeof val);
}